from argparse import ArgumentParser

from zotero2md.zt2md import Zotero2Markdown

if __name__ == "__main__":
    parser = ArgumentParser(description="Generate Markdown files")
    parser.add_argument(
        "zotero_key", help="Zotero API key (visit https://www.zotero.org/settings/keys)"
    )
    parser.add_argument(
        "zotero_user_id",
        help="Zotero User ID (visit https://www.zotero.org/settings/keys)",
    )
    parser.add_argument(
        "--library_type",
        default="user",
        help="Zotero Library type ('user': for personal library (default value), 'group': for a shared library)",
    )
    parser.add_argument(
        "--config_filepath",
        type=str,
        help="Filepath to a .json file containing the path",
    )

    args = vars(parser.parse_args())

    zt = Zotero2Markdown(
        zotero_key=args["zotero_key"],
        zotero_library_id=args["zotero_user_id"],
        zotero_library_type=args["library_type"],
        params_filepath=args.get("config_filepath", None),
        include_annotations=True,
        include_notes=True,
    )

    zt.run_all()
    zt.save_failed_items_to_txt("failed_zotero_items.txt")



from collections import defaultdict
from typing import Dict, List


def group_by_parent_item(annotations: List[Dict]) -> defaultdict:
    annotations_by_parent = defaultdict(list)
    for annot in annotations:
        annotations_by_parent[annot["data"]["parentItem"]].append(annot)
    return annotations_by_parent


def sanitize_tag(tag: str) -> str:
    """Clean tag by replacing empty spaces with underscore.

    Parameters
    ----------
    tag: str

    Returns
    -------
    str
        Cleaned tag

    Examples
    --------
    >>> sanitize_tag(" Machine Learning ")
    "Machine_Learning"

    """
    return tag.strip().replace(" ", "_")


def sanitize_filename(filename: str) -> str:
    """Clean filename
    Replace illegal (or problem-causing) characters with empty spaces, dashes or nothing.
    char_to_replace dictionary contains the mapping.

    Parameters
    ----------
    filename: str
        original filename

    Returns
    -------
    str
        Sanitized filename (without extension)

    Examples
    --------
    >>> sanitize_filename(" abc.pdf")
    "abc"
    >>> sanitize_filename("abc:def?ghi")
    "abc -- ghi"
    >>> sanitize_filename("abc.def")
    "abc def"

    """
    out_filename = filename.strip()
    char_to_replace = {".pdf": " ", ":": " -- ", "/": "-", "?": " ", ".": " "}
    for old, new in char_to_replace.items():
        out_filename = out_filename.replace(old, new)
    return out_filename



from zotero2md.utils import sanitize_tag


def test_sanitize_tag():
    assert sanitize_tag("AB CD") == "AB_CD"
    assert sanitize_tag("AB_-CD") == "AB_-CD"
    assert sanitize_tag(" a b ") == "a_b"



from pathlib import Path

__author__ = "Essi Alizadeh"
__version__ = "0.2.0"

ROOT_DIR: Path = Path(__file__).parent
TOP_DIR: Path = ROOT_DIR.parent

DEFAULT_PARAMS = {
    "convertTagsToInternalLinks": True,
    "doNotConvertFollowingTagsToLink": [],
    "includeHighlightDate": True,
    "hideHighlightDateInPreview": False,
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zotero2md', 'zotero2md.test']

package_data = \
{'': ['*']}

install_requires = \
['Pyzotero>=1.4.26,<2.0.0',
 'SnakeMD>=0.9.3,<0.10.0',
 'markdownify>=0.10.1,<0.11.0']

setup_kwargs = {
    'name': 'zotero2md',
    'version': '0.2.0',
    'description': 'Export Zotero annotations and notes to Markdown files.',
    'long_description': '# Zotero to Markdown\n\nGenerate Markdown files from Zotero annotations and notes. \nWith new [Zotero PDF Reader](https://www.zotero.org/support/pdf_reader_preview), all highlights are saved in the Zotero database.\nThe highlights are NOT saved in the PDF file unless you export the highlights in order to save them.\n\nIf you annotate your files outside the new Zotero PDF reader, this library will not work with your PDF annotations as those are not retrievable from Zotero API.\nIn that case, you may want to use zotfile + mdnotes to extract the annotations and convert them into markdown files.\n\n**_This library is for you if you annotate (highlight + note) using the Zotero\'s PDF reader (including the Zotero iOS)_**\n\n# Installation \nYou can install the library by running \n```shell\npip install zotero2md\n```\n\nNote: If you do not have pip installed on your system, you can follow the instructions [here](https://pip.pypa.io/en/stable/installation/).\n\n# Usage\nSince we have to retrieve the notes from Zotero API, the minimum requirements are:\n* **Zotero API key** [Required]: Create a new Zotero Key from [your Zotero settings](https://www.zotero.org/settings/key)\n* **Zotero personal or group ID** [Required]: \n    * Your **personal library ID** (aka **userID**) can be found [here](https://www.zotero.org/settings/key) next to `Your userID for use in API calls is XXXXXX`.\n    * If you\'re using a **group library**, you can find the library ID by \n        1. Go to `https://www.zotero.org/groups/`\n        2. Click on the interested group.\n        3. You can find the library ID from the URL link that has format like *https://www.zotero.org/groups/<group_id>/group_name*. The number between `/groups/` and `/group_name` is the libarry ID. \n* **Zotero library type** [Optional]: *"user"* (default) if using personal library and *"group"* if using group library.\n\nNote that if you want to retrieve annotations and notes from a group, you should provide the group ID (`zotero_library_id=<group_id>`) and set the library type to group (`zotero_library_type="group"`).\n\n## Approach 1 (Recommended)\nAfter installing the library, open a Python terminal, and then execute the following:\n```python \nfrom zotero2md.zt2md import Zotero2Markdown\n\nzt = Zotero2Markdown(\n    zotero_key="your_zotero_key",  \n    zotero_library_id="your_zotero_id", \n    zotero_library_type="user", # "user" (default) or "group"\n    params_filepath="",  # [Default values provided bellow] # The path to JSON file containing the custom parameters (See Section Custom Output Parameters).\n    include_annotations=True, # Default: True\n    include_notes=True, # Default: True\n)\nzt.run_all()\n```\nJust to make sure that all files are created, you can run `save_failed_items_to_txt()` to ensure that no file was \nwas failed to create. If a file or more failed to create, the filename (item title) and the corresponding Zotero \nitem key will be saved to a txt file. \n```python\nzt.save_failed_items_to_txt("failed_zotero_items.txt")\n```\n\n## Approach 2\nFor this approach, you need to download `output_to_md.py` script. \nRun `python output_to_md.py -h` to get more information about all options. \n```shell\npython zotero2md/output_to_md.py <zotero_key> <zotero_id>\n```\n\nFor instance, assuming zotero_key=abcd and zotero_id=1234, you can simply run the following:\n```shell\npython zotero2md/output_to_md.py abcd 1234\n```\n\n\n## Custom Output Parameters\nYou can change default parameters by passing the `--config_filepath` option with the path to a\nJSON file containing the desired configurations. For instance,\n\n```shell\npython zotero2md/generate.py <zotero_key> <zotero_id> --config_filepath ./sample_params.json\n```\n\n| Parameter                         | type            | default value |\n|-----------------------------------|-----------------|---------------|\n| `convertTagsToInternalLinks`      | bool            | true          |\n| `doNotConvertFollowingTagsToLink` | List of strings | \\[ \\]         |\n| `includeHighlightDate`            | bool            | true          |\n| `hideHighlightDateInPreview`      | bool            | true          |\n\n\nAny parameter in the JSON file will override the default setting. \nIf a parameter is not provided, then the default value will be used. \n\nFor example, if you don\'t want to show the highlight date in the output file, you can simply pass\na JSON file with the following content:\n```json\n{\n  "hideHighlightDateInPreview": false\n}\n```\n\n# Features\n- Generate MD files for all annotations and notes saved in Zotero\n- The ability to convert Zotero tags to internal links (`[[ ]]`) used in many bidirectional MD editors.\n  - You can even pass certain tags that you don\'t want to convert to internal links! (using `doNotConvertFollowingTagsToLink` parameter)\n\n## Quick note\nSince I\'m personally using Obsidian as my markdown editor, there are custom parameters to generate MD files that are consistent with Obsidian and I\'m planning to add more option there. \n\n\n# Roadmap\n- [ ] Update existing annotations and notes\n- [ ] Option to add frontmatter section (particularly useful for Obsidian)\n- [ ] More flexibility in styling the output files \n\n# Request a new feature or report a bug\nFeel free to request a new feature or report a bug in GitHub issue [here](https://github.com/e-alizadeh/Zotero2MD/issues).\n\n## 📫 How to reach me:\n<a href="https://ealizadeh.com" target="_blank"><img alt="Personal Website" src="https://img.shields.io/badge/Personal%20Website-%2312100E.svg?&style=for-the-badge&logoColor=white" /></a>\n<a href="https://www.linkedin.com/in/alizadehesmaeil/" target="_blank"><img alt="LinkedIn" src="https://img.shields.io/badge/linkedin-%230077B5.svg?&style=for-the-badge&logo=linkedin&logoColor=white" /></a>\n<a href="https://medium.com/@ealizadeh" target="_blank"><img alt="Medium" src="https://img.shields.io/badge/medium-%2312100E.svg?&style=for-the-badge&logo=medium&logoColor=white" /></a>\n<a href="https://twitter.com/intent/follow?screen_name=es_alizadeh&tw_p=followbutton" target="_blank"><img alt="Twitter" src="https://img.shields.io/badge/twitter-%231DA1F2.svg?&style=for-the-badge&logo=twitter&logoColor=white" /></a>\n\n<a href="https://www.buymeacoffee.com/ealizadeh" target="_blank"><img src="https://cdn.buymeacoffee.com/buttons/v2/default-blue.png" alt="Buy Me A Coffee" style="height: 60px !important;width: 217px !important;" ></a>',
    'author': 'ealizadeh',
    'author_email': 'hello@ealizadeh.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/e-alizadeh/Zotero2MD',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json
from os import environ
from pathlib import Path
from typing import Dict, List, Tuple, Union

from markdownify import markdownify
from pyzotero.zotero import Zotero
from pyzotero.zotero_errors import ParamNotPassed, UnsupportedParams
from snakemd import Document, MDList, Paragraph

from zotero2md import DEFAULT_PARAMS
from zotero2md.utils import sanitize_filename, sanitize_tag

_OUTPUT_DIR = Path("zotero_output")
_OUTPUT_DIR.mkdir(parents=True, exist_ok=True)


def get_zotero_client(
    library_id: str = None, api_key: str = None, library_type: str = "user"
) -> Zotero:
    """Create a Zotero client object from Pyzotero library

    Zotero userID and Key are available

    Parameters
    ----------
    library_id: str
        If not passed, then it looks for `ZOTERO_LIBRARY_ID` in the environment variables.
    api_key: str
        If not passed, then it looks for `ZOTERO_KEY` in the environment variables.
    library_type: str ['user', 'group']
        'user': to access your Zotero library
        'group': to access a shared group library

    Returns
    -------
    Zotero
        a Zotero client object
    """

    if library_id is None:
        try:
            library_id = environ["ZOTERO_LIBRARY_ID"]
        except KeyError:
            raise ParamNotPassed(
                "No value for library_id is found. "
                "You can set it as an environment variable `ZOTERO_LIBRARY_ID` or use `library_id` to set it."
            )

    if api_key is None:
        try:
            api_key = environ["ZOTERO_KEY"]
        except KeyError:
            raise ParamNotPassed(
                "No value for api_key is found. "
                "You can set it as an environment variable `ZOTERO_KEY` or use `api_key` to set it."
            )

    if library_type is None:
        library_type = environ.get("LIBRARY_TYPE", "user")
    elif library_type not in ["user", "group"]:
        raise UnsupportedParams("library_type value can either be 'user' or 'group'.")

    return Zotero(
        library_id=library_id,
        library_type=library_type,
        api_key=api_key,
    )


class ZoteroItemBase:
    def __init__(
        self,
        zotero_client: Zotero,
        item_key: str,
        params_filepath: Union[str, None] = None,
    ):
        self.zotero = zotero_client
        self.item_key = item_key
        self.item_details = self.zotero.item(self.item_key)
        self.parent_item_key = self.item_details["data"].get("parentItem", None)

        # Load output configurations used for generating markdown files.
        self.md_config = DEFAULT_PARAMS

        if params_filepath:
            with open(Path(params_filepath), "r") as f:
                params = json.load(f)
            self.md_config = {**self.md_config, **params}

    def get_item_metadata(self) -> Dict:
        if self.parent_item_key:
            top_parent_item = self.zotero.item(self.parent_item_key)
            data = top_parent_item["data"]
            metadata = {
                "title": data["title"],
                "date": data["date"],
                "creators": [
                    creator["firstName"] + " " + creator["lastName"]
                    for creator in data["creators"]
                ],
                "tags": data["tags"],
            }
        else:
            data = self.item_details["data"]
            metadata = {
                "title": data["title"],
                "tags": data["tags"],
            }

        return metadata

    def format_tags(self, tags: List[Dict[str, str]]) -> str:
        flattened_tags = [d_["tag"] for d_ in tags]
        if self.md_config["convertTagsToInternalLinks"]:
            return " ".join(
                [
                    f"#{sanitize_tag(tag)}"
                    if tag in self.md_config["doNotConvertFollowingTagsToLink"]  # type: ignore
                    else f"[[{tag}]]"
                    for tag in flattened_tags
                ]
            )
        else:
            return " ".join([f"#{sanitize_tag(tag)}" for tag in flattened_tags])

    def format_metadata(self, metadata: Dict) -> List:
        output: List = []
        authors = metadata.get("creators", None)
        if authors:
            if len(authors) == 1:
                output.append(f"**Author:** {authors[0]}")
            elif 1 < len(authors) <= 5:
                output.append(f"**Authors:** {', '.join(authors)}")
            elif len(authors) > 5:
                output.append(f"**Authors:** {authors[0]} et al.")

        if metadata.get("date", None):
            output.append(f"**Date:** {metadata['date']}")

        tags = metadata.get("tags", None)

        if tags:
            output.append(self.format_tags(tags))
        return output


class ItemAnnotationsAndNotes(ZoteroItemBase):
    def __init__(
        self,
        zotero_client: Zotero,
        item_key: str,
        item_annotations: List[Dict] = None,
        item_notes: List[Dict] = None,
        params_filepath: Union[str, None] = None,
    ):
        """
        Assumptions
        -----------
        item_annotations and item_notes belong to the same Zotero item!

        Parameters
        ----------
        zotero_client
        item_key
        item_annotations
        item_notes
        params_filepath
        """
        super().__init__(zotero_client, item_key, params_filepath)
        self.item_annotations = item_annotations
        self.item_notes = item_notes
        self.failed_item: str = ""

        self.doc = Document(
            name="initial_filename"
        )  # the filename will be updated later!

    def _format_highlighted_date(self, date: str):
        if self.md_config["includeHighlightDate"]:
            if self.md_config["hideHighlightDateInPreview"]:
                return f"<!--(Highlighted on {date})-->"
            else:
                return f"(Highlighted on {date})"
        else:
            return ""

    def format_annotation(self, highlight: Dict) -> Union[Tuple, Paragraph]:
        data = highlight["data"]
        tags = data["tags"]
        # zotero_unique_id = f"(key={highlight['key']}, version={highlight['version']})"
        annot_text = ""
        annot_sub_bullet = []
        if data["annotationType"] == "note":
            annot_text = (
                f"{data['annotationComment']} (Note on *Page {data['annotationPageLabel']}*) "
                + self._format_highlighted_date(data["dateModified"])
                # f"<!---->"
            )
        elif data["annotationType"] == "highlight":
            annot_text = (
                f"{data['annotationText']} (*Page {data['annotationPageLabel']}*) "
                + self._format_highlighted_date(data["dateModified"])
                # f"<!---->"
            )
            if data.get("annotationComment", "") != "":
                annot_sub_bullet.append(f"**Comment**: {data['annotationComment']}")

        if tags:
            annot_sub_bullet.append(self.format_tags(tags))

        return annot_text, MDList(annot_sub_bullet)

    def create_metadata_section(self, metadata: Dict) -> None:
        """Generate the metadata sections (titled "Metadata") containing metadata about the Zotero Item

        Parameters
        ----------
        metadata: Dict
            A dictionary of an item details

        Returns
        -------
        None
        """
        self.doc.add_header(level=1, text="Metadata")
        self.doc.add_element(MDList(self.format_metadata(metadata)))

    def format_note(self, note) -> Tuple[str, MDList]:
        data = note["data"]
        tags = data["tags"]

        md_note = markdownify(data["note"])
        note_sub_bullet = []

        if tags:
            note_sub_bullet.append(self.format_tags(tags))

        return md_note, MDList(note_sub_bullet)

    def create_notes_section(self, notes: List) -> None:
        self.doc.add_header(level=1, text="Notes")
        annots = []
        for note in notes:
            formatted_note = self.format_note(note)
            if isinstance(formatted_note, tuple):
                annots.append(formatted_note[0])
                annots.append(formatted_note[1])
            else:
                annots.append(formatted_note)

        self.doc.add_element(MDList(annots))

    def create_annotations_section(self, annotations: List) -> None:
        """Generate the annotation sections (titled "Highlights")
        In Zotero, an annotation is a highlighted text with the possibility of having related comment and tag(s).
        In addition, a note can also be added to a page without any highlight. This is also considered an annotation.
        The itemType="annotation" in the API response of both scenarios above.

        Parameters
        ----------
        annotations: List[Dict]
            A list containing all annotations of a Zotero Item.

        Returns
        -------
        None
        """
        self.doc.add_header(level=1, text="Highlights")
        annots = []
        for h in annotations:
            formatted_annotation = self.format_annotation(h)
            if isinstance(formatted_annotation, tuple):
                annots.append(formatted_annotation[0])
                annots.append(formatted_annotation[1])
            else:
                annots.append(formatted_annotation)

        self.doc.add_element(MDList(annots))

    def generate_output(self) -> None:
        """Generate the markdown file for a Zotero Item combining metadata, annotations

        Returns
        -------
        List
            Output (filename, item_key) of failed markdown files to compile.

        """
        metadata = self.get_item_metadata()
        title = metadata["title"]

        self.doc.add_header(title, level=1)
        self.create_metadata_section(metadata)
        if self.item_notes:
            self.create_notes_section(self.item_notes)
        if self.item_annotations:
            self.create_annotations_section(self.item_annotations)

        output_filename = sanitize_filename(title) + ".md"
        try:
            with open(_OUTPUT_DIR.joinpath(output_filename), "w+") as f:
                f.write(self.doc.render())
            print(
                f'File "{output_filename}" (item_key="{self.item_key}") was successfully created.\n'
            )
        except:
            msg = f'File "{output_filename}" (item_key="{self.item_key}") is failed to generate.\n'
            print(msg + "SKIPPING...\n")
            self.failed_item = msg


def retrieve_all_annotations(zotero_client: Zotero) -> List[Dict]:
    print(
        "Retrieving ALL annotations from Zotero Database. \nIt may take some time...\n"
    )
    return zotero_client.everything(zotero_client.items(itemType="annotation"))


def retrieve_all_notes(zotero_client: Zotero) -> List[Dict]:
    print("Retrieving ALL notes from Zotero Database. \nIt may take some time...\n")
    return zotero_client.everything(zotero_client.items(itemType="note"))



from typing import Dict, List, Union

from zotero2md.utils import group_by_parent_item
from zotero2md.zotero import (
    ItemAnnotationsAndNotes,
    get_zotero_client,
    retrieve_all_annotations,
    retrieve_all_notes,
)


class Zotero2Markdown:
    def __init__(
        self,
        zotero_key: str,
        zotero_library_id: str,
        zotero_library_type: str = "user",
        params_filepath: str = None,
        include_annotations: bool = True,
        include_notes: bool = True,
    ):

        self.zotero_client = get_zotero_client(
            library_id=zotero_library_id,
            library_type=zotero_library_type,
            api_key=zotero_key,
        )
        self.config_filepath = params_filepath
        self.include_annots = include_annotations
        self.include_notes = include_notes
        self.failed_items: List[str] = []

    def run_all(self, params_filepath: Union[str, None] = None) -> None:
        annots_grouped: Dict = {}
        notes_grouped: Dict = {}
        if self.include_annots:
            annots_grouped = group_by_parent_item(
                retrieve_all_annotations(self.zotero_client)
            )
        if self.include_notes:
            notes_grouped = group_by_parent_item(retrieve_all_notes(self.zotero_client))

        for i, item_key in enumerate(annots_grouped.keys()):
            item = self.zotero_client.item(item_key)
            parent_item_key = item["data"].get("parentItem", None)
            print(f"File {i + 1} of {len(annots_grouped)} is under process ...")
            zotero_item = ItemAnnotationsAndNotes(
                zotero_client=self.zotero_client,
                params_filepath=params_filepath,
                item_annotations=annots_grouped[item_key],
                item_notes=notes_grouped.get(parent_item_key, None),
                item_key=item_key,
            )
            # del notes[parent_item_key]

            if zotero_item.failed_item:
                self.failed_items.append(zotero_item.failed_item)
            else:
                zotero_item.generate_output()

        # for i, item_key in enumerate(notes.keys()):
        #     print(f"File {i + 1} of {len(highlights)} is under process ...")
        #     item = ItemAnnotationsAndNotes(
        #         zotero_client=zotero_client,
        #         params_filepath=params_filepath,
        #         item_annotations=None,
        #         item_notes=notes[item_key],
        #         item_key=item_key,
        #     )
        #     out = item.generate_output()
        #     if out:
        #         failed_files.append(out)
        #     print("\n")

    def save_failed_items_to_txt(self, txt_filepath: str = ""):
        if txt_filepath == "":
            txt_filepath = "failed_zotero_items.txt"

        if self.failed_items:
            print(
                f"\n {len(self.failed_items)} markdown files (with all their annotations and notes) failed to create."
            )
            with open(txt_filepath, "w") as f:
                file_content = "\n".join(self.failed_items)
                f.write(file_content)
            print(f"List of all failed items are saved into {txt_filepath}")
        else:
            print("No failed item")
